use crate::prelude::*;

use serde::{Deserialize, Serialize};
use std::{cell::RefCell, rc::Rc};

pub struct LocalDiscovery {
    api: Rc<crate::API>,
    targets: RefCell<Vec<fleck_core::peer::Peer>>,
}

#[derive(Serialize, Deserialize, Debug)]
struct DiscoveryMsg {
    pkey: fleck_core::crypto::PublicKey,
}

impl fleck_core::msg::MessageParams for DiscoveryMsg {
    const NAME: &'static str = "DiscoveryMsg";
    const ENCRYPTED: bool = false;
}

impl Service for LocalDiscovery {
    fn new(api: Rc<crate::API>) -> Self {
        Self {
            api,
            targets: Default::default(),
        }
    }
    fn setup(self: &Rc<Self>) {
        self.api.with_service(|m: &fleck_core::MessageService| {
            m.add_message_type::<DiscoveryMsg>();
        });
        self.api.channel::<fleck_core::MajorTickChannel>()
            .sub_ref(self, LocalDiscovery::process_major_tick);
        self.api.channel::<fleck_core::msg::MessageChannel<DiscoveryMsg>>()
            .sub_eat(self, LocalDiscovery::discovery);
    }
}

impl LocalDiscovery {
    pub fn add_multicast_peer(&self, peer: fleck_core::peer::Peer) {
        self.targets.borrow_mut().push(peer);
    }

    fn discovery(&self, msg: (fleck_core::msg::Metadata, DiscoveryMsg)) {
        let discovery = msg.1;
        log::trace!("received discovery message");

        // this will automatically ignore self-messages because we already know the pubkey
        self.api.with_service(|ns: &fleck_core::NodeService| {
            if ns.node_by_pubkey(&discovery.pkey).is_none() {
                ns.inform_of_peer(msg.0.peer.unwrap());
                // let node = fleck_core::node::Node::build_with_peer(msg.0.peer);
                // node.set_pubkey(discovery.pkey);
                // nodes.inform_of(node);
            }
        });
    }

    fn process_major_tick(&self, _: &mut ()) {
        for peer in self.targets.borrow().iter() {
            let msg = fleck_core::msg::Message::build(DiscoveryMsg {
                pkey: self.api.with_service(|ns: &fleck_core::NodeService| {
                    ns.self_node().pubkey().unwrap().to_owned()
                }),
            })
            .with_peer(peer.clone());

            self.api.queue::<fleck_core::SendPacketChannel>(msg);
        }
    }
}
