use std::{
    any::Any,
    cell::RefCell,
    collections::HashMap,
    hash::{Hash, Hasher},
    rc::Rc,
};

use serde::{de::DeserializeOwned, Deserialize, Serialize};

use crate::prelude::*;

const MESSAGE_MAGIC: u64 = 0x1234123412341234;

pub trait MessageParams: 'static {
    const NAME: &'static str;
    const ENCRYPTED: bool = true;
    const SIGNED: bool = false;
}

#[derive(Clone, Copy, Debug)]
pub(crate) struct SavedMessageParams {
    pub(crate) name: &'static str,
    pub(crate) encrypted: bool,
    pub(crate) signed: bool,
}

impl SavedMessageParams {
    fn save<M: MessageParams>() -> Self {
        Self {
            name: M::NAME,
            encrypted: M::ENCRYPTED,
            signed: M::SIGNED,
        }
    }
}

fn message_type<M: MessageParams>() -> u16 {
    let mut hasher = std::collections::hash_map::DefaultHasher::new();
    M::NAME.hash(&mut hasher);
    (hasher.finish() & 0xffff) as u16
}

#[derive(Serialize, Deserialize, Debug)]
pub struct MessageContent {
    pub(crate) ty: u16,
    pub(crate) data: Vec<u8>,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct Message {
    magic: u64,
    pub(crate) crypto_header: fleck_core::crypto::PacketHeader,
    pub(crate) content: MessageContent,

    #[serde(skip_serializing)]
    #[serde(skip_deserializing)]
    pub(crate) saved_params: Option<SavedMessageParams>,

    #[serde(skip_serializing)]
    #[serde(skip_deserializing)]
    pub(crate) parsed: Option<Box<dyn Any>>,

    #[serde(skip_serializing)]
    #[serde(skip_deserializing)]
    pub(crate) peer: Option<super::peer::Peer>,
}

#[derive(Default)]
pub struct Metadata {
    pub peer: Option<fleck_core::peer::Peer>,
    pub node: Option<fleck_core::node::Node>,
}

impl Message {
    pub fn build<M: 'static + MessageParams + Serialize>(from: M) -> Self {
        Self {
            magic: MESSAGE_MAGIC,
            crypto_header: Default::default(),
            content: MessageContent {
                ty: message_type::<M>(),
                data: bincode::serialize(&from).expect("couldn't serialize message"),
            },
            saved_params: Some(SavedMessageParams::save::<M>()),
            parsed: Some(Box::new(from)),
            peer: None,
        }
    }

    fn metadata(&self) -> Metadata {
        Metadata {
            peer: self.peer.clone(),
            node: None,
        }
    }

    pub fn with_peer(mut self, peer: fleck_core::peer::Peer) -> Self {
        self.peer = Some(peer);
        self
    }
}

pub struct MessageChannelTag;
pub type MessageChannel<M> = (MessageChannelTag, (Metadata, M));

type Deserializer = dyn Fn(&Message) -> ();

pub struct MessageService {
    api: Rc<crate::API>,
    deser: RefCell<HashMap<u16, Box<Deserializer>>>,
}

impl Service for MessageService {
    fn new(api: Rc<crate::API>) -> Self {
        Self { api, deser: Default::default() }
    }
    fn setup(self: &Rc<Self>) {
        self.api.channel::<fleck_core::ReceivePacketChannel>().sub_opt(
            fleck_core::ReceiveOrder::Parse,
            self,
            Self::parse,
        );
    }
}

impl MessageService {
    pub fn add_message_type<MT: std::fmt::Debug + MessageParams + Serialize + DeserializeOwned>(
        &self
    ) {
        let derived_typeid = message_type::<MT>();
        let api = self.api.clone();
        self.deser.borrow_mut().insert(
            derived_typeid,
            Box::new(
                move |message| match bincode::deserialize::<MT>(&message.content.data) {
                    Ok(content) => {
                        api.queue::<MessageChannel<MT>>((message.metadata(), content));
                    }
                    Err(_) => {
                        log::trace!("Packet failed deserialization step?");
                    }
                },
            ),
        );

        self.api.create_channel::<MessageChannel<MT>>();
    }

    fn parse(&self, msg: Message) -> Option<Message> {
        // try deserializing
        (self.deser.borrow().get(&msg.content.ty)?)(&msg);
        None
    }
}
