use rand::prelude::*;

use std::{
    cell::{Ref, RefCell},
    collections::{HashMap, HashSet},
    rc::Rc,
};

use crate::prelude::*;

use fleck_core::crypto::{Keypair, PublicKey};

pub struct InnerRef<'l, T> {
    r: Ref<'l, Option<T>>,
}

impl<'l, T> InnerRef<'l, T> {
    fn build(from: Ref<'l, Option<T>>) -> Option<Self> {
        if from.is_none() {
            None
        } else {
            Some(Self { r: from })
        }
    }
}

impl<'l, T> AsRef<T> for InnerRef<'l, T> {
    fn as_ref(&self) -> &T {
        self.r.as_ref().unwrap()
    }
}

impl<'l, T> std::ops::Deref for InnerRef<'l, T> {
    type Target = T;
    fn deref(&self) -> &Self::Target {
        self.r.as_ref().unwrap()
    }
}

#[derive(Default)]
pub struct Node {
    peer: Option<fleck_core::peer::Peer>,
    pubkey: RefCell<Option<PublicKey>>,
    keypair: RefCell<Option<Keypair>>,
}

impl Node {
    pub fn build_with_peer(peer: fleck_core::peer::Peer) -> Self {
        Self {
            peer: Some(peer),
            ..Default::default()
        }
    }

    pub fn peer(&self) -> Option<fleck_core::peer::Peer> {
        self.peer.clone()
    }

    pub fn pubkey(&self) -> Option<InnerRef<'_, PublicKey>> {
        InnerRef::build(self.pubkey.borrow())
    }
    pub(crate) fn keypair(&self) -> Option<InnerRef<'_, Keypair>> {
        InnerRef::build(self.keypair.borrow())
    }

    pub fn set_pubkey(&self, pk: PublicKey) {
        let mut pkey = self.pubkey.borrow_mut();
        if pkey.is_some() {
            panic!("Tried to set public key on node that already has a public key!")
        }
        *pkey = Some(pk);
    }

    pub fn set_keypair(&self, kp: Keypair) {
        let mut keyp = self.keypair.borrow_mut();
        if keyp.is_some() {
            panic!("Tried to set keypair on node that already has a keypair!")
        }
        *keyp = Some(kp);
    }

    pub fn gen_keypair(&self) {
        let mut pkey = self.pubkey.borrow_mut();
        let mut keyp = self.keypair.borrow_mut();
        if pkey.is_some() {
            panic!("Tried to generate keypair for node that already has a pubkey!")
        }
        if keyp.is_some() {
            panic!("Tried to generate keypair for node that already has a keypair!")
        }
        let mut csprng = StdRng::from_entropy();
        let kp = Keypair::generate(&mut csprng);
        *pkey = Some(kp.public);
        *keyp = Some(kp);
    }
}

pub struct RegTag {}
pub type NodeRegistrationChannel = (RegTag, Rc<Node>);

pub struct NodeService {
    all_nodes: RefCell<Vec<Rc<Node>>>,
    self_node: Rc<Node>,
}

impl Default for NodeService {
    fn default() -> Self {
        let ret = Self {
            all_nodes: Default::default(),
            self_node: Default::default(),
        };

        ret.all_nodes.borrow_mut().push(ret.self_node.clone());

        ret
    }
}

impl NodeService {
    pub fn self_node(&self) -> &Rc<Node> {
        &self.self_node
    }

    pub fn node_by_pubkey(&self, pubkey: &PublicKey) -> Option<Rc<Node>> {
        self.all_nodes
            .borrow()
            .iter()
            .find(|n| n.pubkey().as_deref() == Some(pubkey))
            .map(|r| r.to_owned())
    }

    // nodes we believe we can communicate with directly
    pub fn direct_neighbours(&self) -> Vec<Rc<Node>> {
        self.all_nodes
            .borrow()
            .iter()
            .filter(|n| n.peer.is_some())
            .map(|r| r.to_owned())
            .collect()
    }

    pub fn inform_of(&self, node: Node) {
        // TODO: check for duplicates here
        self.all_nodes.borrow_mut().push(Rc::new(node));
    }

    pub fn inform_of_peer(&self, peer: fleck_core::peer::Peer) {
        todo!()
    }
}

impl DefaultService for NodeService {
    fn setup(self: &Rc<Self>, api: Rc<crate::API>) {
        api.create_channel::<NodeRegistrationChannel>();
    }
}
