//! Peer: someone we can communicate with directly

use std::{rc::Rc, cell::Cell};

use crate::prelude::*;

pub(crate) enum PeerAddress {
    Stream,
    Udp(std::net::SocketAddr),
    Virtual,
}

pub(crate) struct PeerData {
    pub(crate) io: Rc<dyn super::io::FdHandler>,
    pub(crate) address_info: PeerAddress,
    sequence_number: Cell<u64>,
}

impl PeerData {
    pub(crate) fn new(io: Rc<dyn super::io::FdHandler>, address_info: PeerAddress) -> Self {
        Self {
            io,
            address_info,
            sequence_number: Cell::new(0),
        }
    }

    pub(crate) fn next_sequence(&self) -> u64 {
        let ret = self.sequence_number.get();
        self.sequence_number.set(ret.wrapping_add(1));
        ret
    }

    pub(crate) fn reset_sequence(&self, to: u64) {
        self.sequence_number.set(to);
    }
}

impl std::fmt::Debug for PeerData {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str("PeerData {}")
    }
}

#[derive(Clone, Debug)]
pub struct Peer {
    pub(crate) data: Rc<PeerData>,
}

impl std::hash::Hash for Peer {
    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
        state.write_usize(Rc::as_ptr(&self.data) as usize);
    }
}

impl PartialEq for Peer {
    fn eq(&self, other: &Self) -> bool {
        Rc::as_ptr(&self.data) == Rc::as_ptr(&other.data)
    }
}

impl Eq for Peer {
    fn assert_receiver_is_total_eq(&self) { }
}

#[derive(Default)]
pub struct PeerService {}

impl DefaultService for PeerService {
    fn setup(self: &Rc<Self>, api: Rc<crate::API>) {
        api.channel::<fleck_core::SendPacketChannel>()
            .as_ref()
            .sub_opt(fleck_core::SendOrder::Dispatch, self, Self::dispatch);
    }
}

impl PeerService {
    fn dispatch(
        &self,
        msg: super::msg::Message,
    ) -> Option<super::msg::Message> {
        if msg.peer.is_none() {
            log::error!("Message got to dispatch with no peer selected?");
            return None;
        }
        let peer = msg.peer.clone().unwrap();
        peer.data.io.dispatch(msg);

        None
    }
}
