use std::rc::Rc;

pub mod fleck_core;
mod helper;
pub mod service;

pub mod prelude {
    pub use crate::fleck_core;
    pub use crate::helper::{AsAny, IntoWeak};
    pub use crate::service::{ChannelSpec, Service, DefaultService};
}

use prelude::*;

pub struct API {
    services: std::cell::RefCell<service::ServiceStack>,
}

impl API {
    pub fn new() -> Rc<Self> {
        let res = Rc::new(Self {
            services: Default::default(),
        });

        res.services.borrow_mut().give_api(Rc::downgrade(&res));

        res.add_service::<fleck_core::io::IOService>();
        res.add_service::<fleck_core::CoreInitService>();

        res
    }

    pub fn run(self: &Rc<Self>) {
        self.with_service(|io: &fleck_core::io::IOService| {
            io.run();
        });
    }
}

impl API {
    /// Add a service to the Fleck instance.
    pub fn add_service<S: Service + 'static>(&self)
    {
        self.services.borrow().give_service::<S>();
    }
    /// Access a pre-existing service
    pub fn with_service<S: Service + 'static, R, F: FnOnce(&S) -> R>(&self, f: F) -> R
    {
        self.services.borrow().with_service(f)
    }
    /// Create an event channel from a specification
    pub fn create_channel<CS: service::ChannelSpec>(
        &self,
    ) -> Rc<service::event::Channel<CS::Tag, CS::Data, CS::Priority>> {
        self.services.borrow().event_root().create_channel::<CS>();
        self.channel::<CS>()
    }
    /// Get an existing channel
    pub fn channel<CS: service::ChannelSpec>(
        &self,
    ) -> Rc<service::event::Channel<CS::Tag, CS::Data, CS::Priority>> {
        self.services.borrow().event_root().channel::<CS>()
    }
    /// Queue a message onto a channel
    pub fn queue<CS: service::ChannelSpec>(&self, data: CS::Data) {
        self.services
            .borrow()
            .event_root()
            .channel::<CS>()
            .as_ref()
            .queue(data);
    }
}

pub struct Fleck {
    api: Rc<API>,
}

impl Fleck {
    pub fn new() -> Self {
        Self {
            api: API::new()
        }
    }

    pub fn api(&self) -> Rc<API> {
        self.api.clone()
    }
}
